/*++

Copyright (c) 2008, Microsoft. All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. 
	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 
	Neither the name of the Microsoft nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--*/

#ifndef _VMBUS_API_H_
#define _VMBUS_API_H_

#include "osd.h"

//
// Defines
//

#define MAX_PAGE_BUFFER_COUNT				16
#define MAX_MULTIPAGE_BUFFER_COUNT			32 // 128K


//
// Fwd declarations
//
typedef struct _DRIVER_OBJECT *PDRIVER_OBJECT;
typedef struct _DEVICE_OBJECT *PDEVICE_OBJECT;

//
// Data types
//

#pragma pack(push,1)

// Single-page buffer
typedef struct _PAGE_BUFFER {
	UINT32	Length;
	UINT32	Offset;
	UINT64	Pfn;
} PAGE_BUFFER;

// Multiple-page buffer
typedef struct _MULTIPAGE_BUFFER {
	// Length and Offset determines the # of pfns in the array
	UINT32	Length;
	UINT32	Offset;
	UINT64	PfnArray[MAX_MULTIPAGE_BUFFER_COUNT];
}MULTIPAGE_BUFFER;

//0x18 includes the proprietary packet header
#define MAX_PAGE_BUFFER_PACKET			(0x18 + (sizeof(PAGE_BUFFER) * MAX_PAGE_BUFFER_COUNT))	
#define MAX_MULTIPAGE_BUFFER_PACKET		(0x18 + sizeof(MULTIPAGE_BUFFER))


#pragma pack(pop)

// All drivers
typedef int (*PFN_ON_DEVICEADD)(PDEVICE_OBJECT Device, void* AdditionalInfo);
typedef int (*PFN_ON_DEVICEREMOVE)(PDEVICE_OBJECT Device);
typedef char** (*PFN_ON_GETDEVICEIDS)(void);
typedef void (*PFN_ON_CLEANUP)(PDRIVER_OBJECT Driver);

// Vmbus extensions
//typedef int (*PFN_ON_MATCH)(PDEVICE_OBJECT dev, PDRIVER_OBJECT drv);
//typedef int (*PFN_ON_PROBE)(PDEVICE_OBJECT dev);
typedef int	(*PFN_ON_ISR)(PDRIVER_OBJECT drv);
typedef void (*PFN_ON_DPC)(PDRIVER_OBJECT drv);
typedef void (*PFN_GET_CHANNEL_OFFERS)(void);

typedef PDEVICE_OBJECT (*PFN_ON_CHILDDEVICE_CREATE)(GUID DeviceType, GUID DeviceInstance, void *Context);
typedef void (*PFN_ON_CHILDDEVICE_DESTROY)(PDEVICE_OBJECT Device);
typedef int (*PFN_ON_CHILDDEVICE_ADD)(PDEVICE_OBJECT RootDevice, PDEVICE_OBJECT ChildDevice);
typedef void (*PFN_ON_CHILDDEVICE_REMOVE)(PDEVICE_OBJECT Device);

// Vmbus channel interface
typedef void (*VMBUS_CHANNEL_CALLBACK)(PVOID context);

typedef int	(*VMBUS_CHANNEL_OPEN)(
	PDEVICE_OBJECT		Device,
	UINT32				SendBufferSize,
	UINT32				RecvRingBufferSize,
	PVOID				UserData,
	UINT32				UserDataLen,
	VMBUS_CHANNEL_CALLBACK ChannelCallback,
	PVOID				Context
	);

typedef void (*VMBUS_CHANNEL_CLOSE)(
	PDEVICE_OBJECT		Device
	);

typedef int	(*VMBUS_CHANNEL_SEND_PACKET)(
	PDEVICE_OBJECT		Device, 
	const PVOID			Buffer, 
	UINT32				BufferLen,
	UINT64				RequestId,
	UINT32				Type,
	UINT32				Flags
);

typedef int	(*VMBUS_CHANNEL_SEND_PACKET_PAGEBUFFER)(
	PDEVICE_OBJECT		Device,
	PAGE_BUFFER			PageBuffers[],
	UINT32				PageCount,
	PVOID				Buffer,
	UINT32				BufferLen,
	UINT64				RequestId
	);

typedef int	(*VMBUS_CHANNEL_SEND_PACKET_MULTIPAGEBUFFER)(
	PDEVICE_OBJECT		Device,
	MULTIPAGE_BUFFER	*MultiPageBuffer,
	PVOID				Buffer,
	UINT32				BufferLen,
	UINT64				RequestId
);

typedef int	(*VMBUS_CHANNEL_RECV_PACKET)(
	PDEVICE_OBJECT		Device,
	PVOID				Buffer,
	UINT32				BufferLen,
	UINT32*				BufferActualLen,
	UINT64*				RequestId
	);

typedef int	(*VMBUS_CHANNEL_RECV_PACKET_PAW)(
	PDEVICE_OBJECT		Device,
	PVOID				Buffer,
	UINT32				BufferLen,
	UINT32*				BufferActualLen,
	UINT64*				RequestId
	);

typedef int	(*VMBUS_CHANNEL_ESTABLISH_GPADL)(
	PDEVICE_OBJECT		Device,
	PVOID				Buffer,	// from kmalloc()
	UINT32				BufferLen,		// page-size multiple
	UINT32*				GpadlHandle
	);

typedef int	(*VMBUS_CHANNEL_TEARDOWN_GPADL)(
	PDEVICE_OBJECT		Device,
	UINT32				GpadlHandle
	);


typedef struct _PORT_INFO {
	UINT32		InterruptMask;
	UINT32		ReadIndex;
	UINT32		WriteIndex;
	UINT32		BytesAvailToRead;
	UINT32		BytesAvailToWrite;
} PORT_INFO;


typedef struct _DEVICE_INFO {
	UINT32		ChannelId;
	UINT32		ChannelState;
	GUID		ChannelType;
	GUID		ChannelInstance;

	UINT32						MonitorId;
	UINT32						ServerMonitorPending;
	UINT32						ServerMonitorLatency;
	UINT32						ServerMonitorConnectionId;
	UINT32						ClientMonitorPending;
	UINT32						ClientMonitorLatency;
	UINT32						ClientMonitorConnectionId;

	PORT_INFO	Inbound;
	PORT_INFO	Outbound;
} DEVICE_INFO;

typedef void (*VMBUS_GET_CHANNEL_INFO)(PDEVICE_OBJECT Device, DEVICE_INFO* DeviceInfo);

typedef struct _VMBUS_CHANNEL_INTERFACE {
	VMBUS_CHANNEL_OPEN							Open;
	VMBUS_CHANNEL_CLOSE							Close;
	VMBUS_CHANNEL_SEND_PACKET					SendPacket;
	VMBUS_CHANNEL_SEND_PACKET_PAGEBUFFER		SendPacketPageBuffer;
	VMBUS_CHANNEL_SEND_PACKET_MULTIPAGEBUFFER	SendPacketMultiPageBuffer;
	VMBUS_CHANNEL_RECV_PACKET					RecvPacket;
	VMBUS_CHANNEL_RECV_PACKET_PAW				RecvPacketRaw;
	VMBUS_CHANNEL_ESTABLISH_GPADL				EstablishGpadl;
	VMBUS_CHANNEL_TEARDOWN_GPADL				TeardownGpadl;
	VMBUS_GET_CHANNEL_INFO						GetInfo;
} VMBUS_CHANNEL_INTERFACE;

typedef void (*VMBUS_GET_CHANNEL_INTERFACE)(VMBUS_CHANNEL_INTERFACE *Interface);

// Base driver object
typedef struct _DRIVER_OBJECT {
	const char*				name;
	GUID					deviceType; // the device type supported by this driver

	PFN_ON_DEVICEADD		OnDeviceAdd;
	PFN_ON_DEVICEREMOVE		OnDeviceRemove;
	PFN_ON_GETDEVICEIDS		OnGetDeviceIds; // device ids supported by this driver
	PFN_ON_CLEANUP			OnCleanup;

	VMBUS_CHANNEL_INTERFACE VmbusChannelInterface;
} DRIVER_OBJECT;


// Base device object
typedef struct _DEVICE_OBJECT {
	DRIVER_OBJECT*		Driver;		// the driver for this device		
	char				name[64];
	GUID				deviceType; // the device type id of this device
	GUID				deviceInstance; // the device instance id of this device
	void*				context;
	void*				Extension;		// Device extension;
} DEVICE_OBJECT;


// Vmbus driver object
typedef struct _VMBUS_DRIVER_OBJECT {
	DRIVER_OBJECT		Base; // !! Must be the 1st field !!
		
	// Set by the caller
	PFN_ON_CHILDDEVICE_CREATE	OnChildDeviceCreate;
	PFN_ON_CHILDDEVICE_DESTROY	OnChildDeviceDestroy;
	PFN_ON_CHILDDEVICE_ADD		OnChildDeviceAdd;
	PFN_ON_CHILDDEVICE_REMOVE	OnChildDeviceRemove;

	// Set by the callee
	//PFN_ON_MATCH		OnMatch;
	//PFN_ON_PROBE		OnProbe;
	PFN_ON_ISR				OnIsr;
	PFN_ON_DPC				OnMsgDpc;
	PFN_ON_DPC				OnEventDpc;
	PFN_GET_CHANNEL_OFFERS	GetChannelOffers;

	VMBUS_GET_CHANNEL_INTERFACE GetChannelInterface;
	VMBUS_GET_CHANNEL_INFO		GetChannelInfo;		
} VMBUS_DRIVER_OBJECT;


//
// Interface
//
int 
VmbusInitialize(
	DRIVER_OBJECT* drv
	);

#endif // _VMBUS_API_H_
