/*
 *  Copyright 2009 Citrix Systems, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  For clarity, the licensor of this program does not intend that a
 *  "derivative work" include code which compiles header information from
 *  this program. 
 *
 */

/*++

Name:	inputvsc_drv.c

Desc:	Open-source input vsc driver

--*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/workqueue.h>
#include <linux/input.h>
#if defined(KERNEL_2_6_16) || defined(KERNEL_2_6_18)
#include "../hid/hid.h"
#else
#include <linux/hid.h>
#endif
#include <linux/hiddev.h>

#include "logging.h"
#include "vmbus.h"

#include "InputVscApi.h"

//
// #defines
//

//
// Data types
//

struct input_device_context {
	struct device_context	*device_ctx; // point back to our device context
	struct hid_device		*hid_device;
	INPUT_DEVICE_INFO		device_info;
	int						connected;
};

struct inputvsc_driver_context {
	// !! These must be the first 2 fields !!
	struct driver_context	drv_ctx;
	INPUTVSC_DRIVER_OBJECT	drv_obj;
};

static void inputvsc_deviceinfo_callback(DEVICE_OBJECT *dev, INPUT_DEVICE_INFO* info);
static void inputvsc_inputreport_callback(DEVICE_OBJECT *dev, void* packet, UINT32 len);
static void inputvsc_reportdesc_callback(DEVICE_OBJECT *dev, void* packet, UINT32 len);
static int inputvsc_probe(struct device *device);
static int inputvsc_remove(struct device *device);
static int inputvsc_hid_open(struct hid_device *);
static void inputvsc_hid_close(struct hid_device *);



// The one and only one
static struct inputvsc_driver_context g_inputvsc_drv;
	



/*++

Name:	inputvsc_drv_init()

Desc:	Driver initialization.

--*/
int inputvsc_drv_init(PFN_DRIVERINITIALIZE pfn_drv_init)
{
	int ret=0;
	INPUTVSC_DRIVER_OBJECT *input_drv_obj=&g_inputvsc_drv.drv_obj;
	struct driver_context *drv_ctx=&g_inputvsc_drv.drv_ctx;

	DPRINT_ENTER(INPUTVSC_DRV);

	vmbus_get_interface(&input_drv_obj->Base.VmbusChannelInterface);

	input_drv_obj->OnDeviceInfo = inputvsc_deviceinfo_callback;
	input_drv_obj->OnInputReport = inputvsc_inputreport_callback;
	input_drv_obj->OnReportDescriptor = inputvsc_reportdesc_callback;

	// Callback to client driver to complete the initialization
	pfn_drv_init(&input_drv_obj->Base);

	drv_ctx->driver.name = input_drv_obj->Base.name;
	memcpy(&drv_ctx->class_id, &input_drv_obj->Base.deviceType, sizeof(GUID));

#if defined(KERNEL_2_6_5) || defined(KERNEL_2_6_9)
	drv_ctx->driver.probe = inputvsc_probe;
	drv_ctx->driver.remove = inputtvsc_remove;
#else
	drv_ctx->probe = inputvsc_probe;
	drv_ctx->remove = inputvsc_remove;
#endif 

	// The driver belongs to vmbus
	vmbus_child_driver_register(drv_ctx);

	DPRINT_EXIT(INPUTVSC_DRV);

	return ret;
}


static int inputvsc_drv_exit_cb(struct device *dev, void *data)
{
	struct device **curr = (struct device **)data;
	*curr = dev;
	return 1; // stop iterating
}

/*++

Name:	inputvsc_drv_exit()

Desc:	

--*/
void inputvsc_drv_exit(void)
{
	INPUTVSC_DRIVER_OBJECT *inputvsc_drv_obj=&g_inputvsc_drv.drv_obj;
	struct driver_context *drv_ctx=&g_inputvsc_drv.drv_ctx;

	struct device *current_dev=NULL;

#if defined(KERNEL_2_6_5) || defined(KERNEL_2_6_9)
#define driver_for_each_device(drv, start, data, fn) \
	struct list_head *ptr, *n; \
	list_for_each_safe(ptr, n, &((drv)->devices)) {\
		struct device *curr_dev;\
		curr_dev = list_entry(ptr, struct device, driver_list);\
		fn(curr_dev, data);\
	}
#endif // KERNEL_2_6_9

	DPRINT_ENTER(INPUTVSC_DRV);

	while (1)
	{
		current_dev = NULL;

		// Get the device
		driver_for_each_device(&drv_ctx->driver, NULL, (void*)&current_dev, inputvsc_drv_exit_cb);

		if (current_dev == NULL)
			break;

		// Initiate removal from the top-down
		device_unregister(current_dev);
	}

	if (inputvsc_drv_obj->Base.OnCleanup)
		inputvsc_drv_obj->Base.OnCleanup(&inputvsc_drv_obj->Base);

	vmbus_child_driver_unregister(drv_ctx);

	DPRINT_EXIT(INPUTVSC_DRV);

	return;
}

/*++

Name:	inputvsc_probe()

Desc:	Add a new device for this driver

--*/
static int inputvsc_probe(struct device *device)
{
	int ret=0;

	struct driver_context *driver_ctx = driver_to_driver_context(device->driver);
	struct inputvsc_driver_context *inputvsc_drv_ctx = (struct inputvsc_driver_context*)driver_ctx;
	INPUTVSC_DRIVER_OBJECT* inputvsc_drv_obj = &inputvsc_drv_ctx->drv_obj;

	struct device_context *device_ctx = device_to_device_context(device);
	DEVICE_OBJECT* device_obj = &device_ctx->device_obj;
	struct input_device_context *input_dev_ctx;

	DPRINT_ENTER(INPUTVSC_DRV);

	input_dev_ctx = kmalloc(sizeof(struct input_device_context), GFP_KERNEL);

	device->driver_data = input_dev_ctx;

	// Call to the vsc driver to add the device
	ret = inputvsc_drv_obj->Base.OnDeviceAdd(device_obj, NULL);
	if (ret != 0)
	{
		DPRINT_ERR(INPUTVSC_DRV, "unable to add input vsc device");
		DPRINT_EXIT(INPUTVSC_DRV);

		return -1;
	}

	
	DPRINT_EXIT(INPUTVSC_DRV);

	return 0;
}


/*++

Name:	inputvsc_remove()

Desc:	Callback when our device is removed

--*/
static int inputvsc_remove(struct device *device)
{
	int ret=0;

	struct driver_context *driver_ctx = driver_to_driver_context(device->driver);
	struct inputvsc_driver_context *inputvsc_drv_ctx = (struct inputvsc_driver_context*)driver_ctx;
	INPUTVSC_DRIVER_OBJECT* inputvsc_drv_obj = &inputvsc_drv_ctx->drv_obj;

	struct device_context *device_ctx = device_to_device_context(device);
	DEVICE_OBJECT* device_obj = &device_ctx->device_obj;
	struct input_device_context *input_device_ctx = (struct input_device_context *)device_ctx->device.driver_data ;


	DPRINT_ENTER(INPUTVSC_DRV);

	if (input_device_ctx->connected)
	{
		hidinput_disconnect(input_device_ctx->hid_device);
		input_device_ctx->connected = 0;
	}

	if (!inputvsc_drv_obj->Base.OnDeviceRemove)
	{
		DPRINT_EXIT(INPUTVSC_DRV);
		return -1;
	}

	// Call to the vsc driver to let it know that the device is being removed
	ret = inputvsc_drv_obj->Base.OnDeviceRemove(device_obj);
	if (ret != 0)
	{
		// TODO:
		DPRINT_ERR(INPUTVSC_DRV, "unable to remove vsc device (ret %d)", ret);
	}

	kfree(input_device_ctx);

	DPRINT_EXIT(INPUTVSC_DRV);

	return ret;
}

void inputvsc_deviceinfo_callback(DEVICE_OBJECT *dev, INPUT_DEVICE_INFO* info)
{
	struct device_context *device_ctx = to_device_context(dev);
	struct input_device_context *input_device_ctx = (struct input_device_context *)device_ctx->device.driver_data ;

	memcpy(&input_device_ctx->device_info, info, sizeof(INPUT_DEVICE_INFO));
		
	DPRINT_INFO(INPUTVSC_DRV, "inputvsc_deviceinfo_callback()");

}

void inputvsc_inputreport_callback(DEVICE_OBJECT *dev, void* packet, UINT32 len)
{
	int ret=0;

	struct device_context *device_ctx = to_device_context(dev);
	struct input_device_context *input_device_ctx = (struct input_device_context *)device_ctx->device.driver_data ;

	DPRINT_ENTER(INPUTVSC_DRV);

	ret= hid_input_report(input_device_ctx->hid_device, HID_INPUT_REPORT, packet, len, 1);
		
	DPRINT_DBG(INPUTVSC_DRV, "hid_input_report (ret %d)", ret);

	DPRINT_EXIT(INPUTVSC_DRV);
}
#define NBITS(x) (((x)/BITS_PER_LONG)+1)

void inputvsc_reportdesc_callback(DEVICE_OBJECT *dev, void* packet, UINT32 len)
{
	struct device_context *device_ctx = to_device_context(dev);
	struct input_device_context *input_device_ctx = (struct input_device_context *)device_ctx->device.driver_data ;
	struct hid_device *hid_dev;

	DPRINT_ENTER(INPUTVSC_DRV);

	//hid_debug = -1;

	hid_dev = hid_parse_report(packet, len);
	
	if (hid_dev)
	{
		DPRINT_INFO(INPUTVSC_DRV, "hid_device created");

		hid_dev->hid_open =inputvsc_hid_open;
		hid_dev->hid_close =inputvsc_hid_close;

		hid_dev->bus =  0x06;// BUS_VIRTUAL;
		hid_dev->vendor = input_device_ctx->device_info.VendorID;
		hid_dev->product = input_device_ctx->device_info.ProductID;
		hid_dev->version = input_device_ctx->device_info.VersionNumber;
		hid_dev->dev = &device_ctx->device;

		sprintf(hid_dev->name, "%s", input_device_ctx->device_info.Name);

		if (!hidinput_connect(hid_dev))
		{
			hid_dev->claimed |= HID_CLAIMED_INPUT;
                
			input_device_ctx->connected = 1;

			DPRINT_INFO (INPUTVSC_DRV, "HID device claimed by input\n");

		}

		if (!hid_dev->claimed) {
                DPRINT_ERR (INPUTVSC_DRV, "HID device not claimed by input or hiddev\n");
        }

		input_device_ctx->hid_device = hid_dev;
	}
	DPRINT_EXIT(INPUTVSC_DRV);

}

int inputvsc_hid_open(struct hid_device *hid)
{
	DPRINT_ENTER(INPUTVSC_DRV);
	DPRINT_EXIT(INPUTVSC_DRV);

	return 0;
}

void inputvsc_hid_close(struct hid_device *hid)
{
	DPRINT_ENTER(INPUTVSC_DRV);
	DPRINT_EXIT(INPUTVSC_DRV);
}

MODULE_LICENSE("GPL");

static int __init inputvsc_init(void)
{
	int ret;

	DPRINT_ENTER(INPUTVSC_DRV);
	DPRINT_INFO(INPUTVSC_DRV, "Inputvsc initializing....");

	ret = inputvsc_drv_init(InputVscInitialize);

	DPRINT_EXIT(INPUTVSC_DRV);

	return ret;
}

static void __exit inputvsc_exit(void)
{
	DPRINT_ENTER(INPUTVSC_DRV);

	inputvsc_drv_exit();

	DPRINT_ENTER(INPUTVSC_DRV);
}

module_init(inputvsc_init);
module_exit(inputvsc_exit);

// eof
